@@ -66,12 +66,11 @@ enum xorg::testing::Process::State xorg::testing::Process::GetState() {
   return d_->state;
 }
 
-void xorg::testing::Process::Start(const std::string &program, const std::vector<std::string> &argv) {
+pid_t xorg::testing::Process::Fork() {
   if (d_->pid != -1)
-    throw std::runtime_error("Attempting to start an already started process");
+    throw std::runtime_error("A process may only be forked once");
 
   d_->pid = fork();
-
   if (d_->pid == -1) {
     d_->state = ERROR;
     throw std::runtime_error("Failed to fork child process");
@@ -85,17 +84,38 @@ void xorg::testing::Process::Start(const std::string &program, const std::vector
 #ifdef __linux
     prctl(PR_SET_PDEATHSIG, SIGTERM);
 #endif
+  }
 
+  d_->state = RUNNING;
+  return d_->pid;
+}
+
+void xorg::testing::Process::Start(const std::string &program, const std::vector<std::string> &argv) {
+  if (d_->pid == -1)
+    d_->pid = Fork();
+  else if (d_->pid > 0)
+    throw std::runtime_error("Start() may only be called on the child process");
+
+  if (d_->pid == 0) { /* Child */
     std::vector<char*> args;
     std::vector<std::string>::const_iterator it;
 
+    char *valgrind = getenv("XORG_GTEST_USE_VALGRIND");
+    if (valgrind) {
+      char *tok = strtok(valgrind, " ");
+      while(tok) {
+        args.push_back(strdup(tok));
+        tok = strtok(NULL, " ");
+      }
+    }
+
     args.push_back(strdup(program.c_str()));
 
     for (it = argv.begin(); it != argv.end(); it++)
       args.push_back(strdup(it->c_str()));
     args.push_back(NULL);
 
-    execvp(program.c_str(), &args[0]);
+    execvp(args[0], &args[0]);
 
     d_->state = ERROR;
     throw std::runtime_error("Failed to start process");
@@ -127,24 +147,44 @@ void xorg::testing::Process::Start(const std::string& program, ...) {
 }
 
 bool xorg::testing::Process::WaitForExit(unsigned int timeout) {
-  for (unsigned int i = 0; i < timeout * 100; i++) {
-    int status;
-    int pid = waitpid(Pid(), &status, WNOHANG);
+  sigset_t sig_mask, old_mask;
+  sigemptyset(&sig_mask);
+  sigaddset(&sig_mask, SIGCHLD);
 
-    if (pid == Pid())
-      return true;
-    else if (pid == -1)
-      return errno == ECHILD;
+  if (sigprocmask(SIG_BLOCK, &sig_mask, &old_mask) == 0) {
+    struct timespec sig_timeout = {timeout / 1000,
+                                   (timeout % 1000) * 1000000L};
 
-      usleep(10);
+    if (sigtimedwait(&sig_mask, NULL, &sig_timeout) != SIGCHLD && errno != EAGAIN)
+      usleep(timeout * 1000);
+
+    if (!sigismember(&sig_mask, SIGCHLD)) {
+      if (sigprocmask(SIG_UNBLOCK, &sig_mask, NULL) == -1)
+        std::cout << "WARNING: Failed to unblock SIGCHLD. Tests may behave funny.\n";
+    }
+
+  } else { /* oops, can't wait for SIGCHLD, sleep instead */
+    usleep(timeout * 1000);
   }
 
-  return false;
+  int status;
+  int pid = waitpid(Pid(), &status, WNOHANG);
+  if (pid == Pid()) {
+    if (WIFEXITED(status)) {
+      d_->state = WEXITSTATUS(status) ? FINISHED_FAILURE : FINISHED_SUCCESS;
+    } else if (WIFSIGNALED(status)) {
+      d_->state = FINISHED_FAILURE;
+    }
+    return true;
+  } else {
+    /* prevent callers from getting odd erros if they check for errno */
+    if (pid == 0)
+      errno = 0;
+    return (pid == -1 && errno == ECHILD);
+  }
 }
 
 bool xorg::testing::Process::KillSelf(int signal, unsigned int timeout) {
-  bool wait_success = true;
-
   enum State state = GetState();
   switch (state) {
     case FINISHED_SUCCESS:
@@ -169,12 +209,18 @@ bool xorg::testing::Process::KillSelf(int signal, unsigned int timeout) {
       d_->state = ERROR;
       return false;
     }
-    if (timeout > 0)
+    if (timeout > 0) {
+      bool wait_success = true;
+
       wait_success = WaitForExit(timeout);
+      if (wait_success)
+        d_->pid = -1;
+      return wait_success;
+    }
     d_->pid = -1;
   }
   d_->state = TERMINATED;
-  return wait_success;
+  return true;
 }
 
 bool xorg::testing::Process::Terminate(unsigned int timeout) {
